import inspect
import numpy as np

def f(x):
    pass
function_type = type(f)

def checkargs(function):
    def _f(*arguments):
        for index, argument in enumerate(inspect.getfullargspec(function)[0]):
            
            if argument == 'self':
                continue

            if not isinstance(arguments[index], function.__annotations__[argument]):
                raise TypeError("{} is not of type {}".format(arguments[index], function.__annotations__[argument]))
        return function(*arguments)
    _f.__doc__ = function.__doc__
    return _f


def get_stats(optimizer, NIter, OptArgs,**ExternOptArgs) -> dict:
    """
    ------------------------------------------------------
    Description:
        Return a dictionary with information about how the opmizer algorithm
        perform after N evaluations applied to opmizer function.

        The dictionary has the mean solution obtained, standard desviation, worst and 
        best solution.

    Arguments:
        - optimizer: optimization class.
        - NIter: evaluation number, which is the number of times to 
        applied class.optimize(args).
        - OptArgs: Arguments necessary to perform.
        - ExternOptArgs(optional): additional arguments.
    ------------------------------------------------------
    """
    f_ = []
    x_ = []
    
    for i in range(NIter):
        optimizer.optimize(*OptArgs,**ExternOptArgs)
        f_.append(optimizer.logger['best_f'])
        x_.append(optimizer.logger['best_individual'])
        
    IndWorst = np.argmax(f_)
    IndBest = np.argmin(f_)
    stats_ = {"Worst solution":{} , "Best solution":{}}
    
    stats_["Worst solution"]["x"] = x_[IndWorst]
    stats_["Best solution"]["x"] = x_[IndBest]
    
    stats_["Worst solution"]["f"] = f_[IndWorst]
    stats_["Best solution"]["f"] = f_[IndBest]
    
    stats_["Mean"] = np.mean(f_)
    stats_["Standard deviation"] = np.std(f_)
    stats_['Median'] = np.median(f_)

    return stats_

# -----------------------------------------------------------------------
#                   Configuration by Evolutionary computing 
# -----------------------------------------------------------------------

class OptimizerConfig:
    def __init__(self):
        self.cross_op          = None
        self.mutation_op       = None
        self.survivor_selector = None 
        self.fixer             = None

    def cross(self, crossover_: function_type):
        self.cross_op = crossover_
        return self

    def mutate(self, mutate_ : function_type):
        self.mutation_op = mutate_
        return self
    
    def survivor_selection(self, survivor_function: function_type):
        self.survivor_selector = survivor_function
        return self

    def fixer_invalide_solutions(self, fixer_function: function_type):
        self.fixer = fixer_function
        return self
    #-----------------------------------------------------
                    #Private function.
    #----------------------------------------------------- 
    def default_printable(self) -> str:
        printable = "--------------------------------\n\tConfiguration\n--------------------------------\n"
        if self.cross_op != None:
            printable += "Crossover operator: "+self.cross_op.__doc__ + '\n'
        if self.mutation_op != None:
            printable += "Mutation operator: "+ self.mutation_op.__doc__ + '\n'
        if self.survivor_selector != None:
            printable += "Survivor selection: " + self.survivor_selector.__doc__ + '\n'
        if self.fixer != None:
            printable += "Fixer: " + self.fixer.__doc__ + '\n'
        return printable

class GeneticConfig(OptimizerConfig):
    def __init__(self):
        super().__init__()
        self.parent_selector = None

    def __str__(self):
        printable = self.default_printable()
        if self.parent_selector != None:
            printable += "Parent selection: "+self.parent_selector.__doc__ + '\n'
        printable+="\n--------------------------------"
        return printable
    
    def parent_selection(self, parent_function: function_type):
        self.parent_selector = parent_function
        return self

class EvolutionStrategyConfig(OptimizerConfig):
    def __init__(self):
        super().__init__()
        self.adaptive_crossover_op = None
        self.adaptive_mutation_op  = None

    def __str__(self):
        printable = self.default_printable()        
        if self.adaptive_crossover_op != None:
            printable += "Adaptive crossover: "+ self.adaptive_crossover_op.__doc__ + '\n'
        if self.adaptive_mutation_op != None:
            printable += "Adaptive mutation: " + self.adaptive_mutation_op.__doc__ + '\n'
        printable+="\n--------------------------------"
        return printable

    def adaptive_crossover(self, adaptive_crossover_function : function_type):
        self.adaptive_crossover_op = adaptive_crossover_function
        return self
    
    def adaptive_mutation(self, adaptive_mutation_function: function_type):
        self.adaptive_mutation_op = adaptive_mutation_function
        return self

class EvolutionaryProgrammingConfig(OptimizerConfig):
    def __init__(self):
        super().__init__()
        self.adaptive_mutation_op  = None

    def __str__(self):
        printable = self.default_printable()        
        if self.adaptive_mutation_op != None:
            printable += "Adaptive mutation: " + self.adaptive_mutation_op.__doc__ + '\n'
        printable+="\n--------------------------------"
        return printable    

    def adaptive_mutation(self, adaptive_mutation_function: function_type):
        self.adaptive_mutation_op = adaptive_mutation_function
        return self         

class ContinuosFixer:
    def __init__(self, bounds: list):
        self.Bounds = bounds
        self.__doc__ = "continuos"

    def __call__(self, X_: np.ndarray, ind: int):
        return np.clip(X_[ind], self.Bounds[0], self.Bounds[1])

class NoneFixer:
    def __init__(self):
        self.__doc__ = "None"

    def __call__(self, X_: np.ndarray, ind: int):
        return X_
        


import numpy as np 
from numba import jit,prange
from copy import deepcopy

__all__ = ['proporcional_sampler', 'roulette_sampler', 'stochastic_universal_sampler',\
           'deterministic_sampler', 'tournament_sampler', 'merge_selector', 'replacement_selector',\
            'proporcional_sampling', 'roulette_sampling', 'stochastic_universal_sampling',\
            'deterministic_sampling', 'tournament_sampling']

def get_expected_values(aptitude: np.array) -> np.array:
    averageAptitude = np.sum(aptitude)
    N = len(aptitude)
    return aptitude/averageAptitude * N 

"""
---------------------------------------------------------------------------------
                            |   Genetic Algorithms.   |
---------------------------------------------------------------------------------
"""


"""
---------------------------------------------------------------------------------
                                Parent selection.
---------------------------------------------------------------------------------
"""
@jit(nopython=True,parallel=False)
def proporcional_sampling(expectedVals: np.ndarray) -> np.ndarray:
    return np.arange(len(expectedVals))

@jit(nopython=True,parallel=True)
def roulette_sampling(expectedVals : np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        selection algorithm O(nlogn) for genetic algorithm. 
    Arguments:
        probabilities: 1-D numpy array which is the probability of every 
        individual based on the aptitude.
    About:
        If you want more information, check:
        
    ------------------------------------------------------
    """
    expectedVals = get_expected_values(expectedVals)
    N = len(expectedVals)
    # expectedVals_ = np.copy(expectedVals)

    expectedCumulative = np.cumsum(expectedVals)
    r = np.random.uniform(0.0,expectedCumulative[-1], N)
    return np.searchsorted(expectedCumulative, r) #sample
    
@jit(nopython=False, parallel=False)   
def stochastic_universal_sampling(expectedVals : np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------ 
    Arguments:
        probabilities: 1-D numpy array which is the probability of every 
        individual based on the aptitude.
    About:
        If you want more information, check:
        
    ------------------------------------------------------
    """
    expectedVals = get_expected_values(expectedVals)

    N = len(expectedVals)
    # expectedVals_ =np.copy(expectedVals)

    r = np.random.uniform(0,1)
    currentSum = 0
    sample = []
    for i in range(N):
        currentSum += expectedVals[i]
        while r < currentSum:
            sample.append(i)
            r += 1
    return np.array(sample)

@jit(nopython=False, parallel=True)
def deterministic_sampling(expectedVals : np.ndarray) -> np.ndarray:
    expectedVals = get_expected_values(expectedVals)

    N = len(expectedVals)
    # expectedVals_ = np.copy(expectedVals)

    integer_part = np.array(expectedVals, dtype=int)
    indices = np.arange(N)
    sample = np.repeat(indices,integer_part) 
    
    if(len(sample) < N):
        float_part = expectedVals - integer_part
        ind = np.argpartition(float_part,len(sample) - N)[len(sample) - N:]
        sample = np.concatenate((sample, indices[ind]))
    
    return sample[:N]

@jit(nopython=False, parallel=True)
def tournament_sampling( expectedVals : np.ndarray, chunks : int=2 , prob: float=1.0) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        selection algorithm O(NP) for genetic algorithm where N is the population
        size and P is the number of chunks. 
    Arguments:
        probabilities: 1-D numpy array which is the probability of every 
        individual based on the aptitude.
        chunks: Integer value which is the number of chunks (by default is 2).
        prob: Float value which means if choose the max or min value in every chunk
        (by default is 1). 
        About:
            If you want more information, check:
            
    ------------------------------------------------------
    """ 
    expectedVals = get_expected_values(expectedVals)
    N=len(expectedVals)
    indices = np.arange(N)
    chunks_ = np.ceil(N/chunks)

    sample = []
    ind = -1
    for tournament in range(chunks):
        groups = np.array_split(np.random.permutation(indices), chunks_)
        for i in prange(len(groups)):
            if prob >= np.random.rand():
                ind = np.argmax(expectedVals[groups[i]])
            else:
                ind = np.argmin(expectedVals[groups[i]])
            sample.append(groups[i][ind])

    return np.array(sample)
    

"""
---------------------------------------------------------------------------------
                                Survivor selection.
---------------------------------------------------------------------------------
"""
def get_lowest_indices(value: np.ndarray, m: int) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Return m indices with the lowest fitness.
    Arguments:
        value: float array.
        m: int element.
    ------------------------------------------------------
    """
    pseudo_sorted_array = np.argpartition(value, m)    
    return pseudo_sorted_array[:m]



"""
---------------------------------------------------------------------------------
                                Classes
---------------------------------------------------------------------------------
"""

"""
----------------------------
    Parent Selection.
----------------------------
"""


class proporcional_sampler:
    def __init__(self, transform=None):
        self.__doc__ = "Proporcional sampling"

    def __call__(self,population_f: np.ndarray) -> np.ndarray:
        return proporcional_sampling(population_f)

class roulette_sampler:
    def __init__(self, transform=None):
        self.__doc__ = "Roulette sampling"
        self.transform = transform

    def __call__(self,population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        if self.transform != None:
            vals = self.transform(vals)
        
        return roulette_sampling(vals)

class stochastic_universal_sampler:
    def __init__(self, transform=None):
        self.__doc__   = "Stochastic universal sampling"
        self.transform = transform

    def __call__(self,population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        if self.transform != None:
            vals = self.transform(vals)
        
        return stochastic_universal_sampling(vals)

class deterministic_sampler:
    def __init__(self, transform=None):
        self.__doc__   = "Deterministic sampling"
        self.transform = transform

    def __call__(self,population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        if self.transform != None:
            vals = self.transform(vals)
        
        return deterministic_sampling(vals)

class tournament_sampler:
    def __init__(self, transform=None, chunks_ : int=2 , prob_: float=1.0):  
        self.transform = transform
        self.chunks = chunks_
        self.prob = prob_
        self.__doc__ = "Tournament sampling\n\t Arguments:\n\t\t-Chunks: {}\n\t\t-prob: {}".format(self.chunks,self.prob)

    def __call__(self,population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        if self.transform != None:
            vals = self.transform(vals)
        
        return tournament_sampling( vals,\
                                    self.chunks,\
                                    self.prob)

"""
----------------------------
    Survivor Selection.
----------------------------
"""

class merge_selector:

    def __init__(self):
        self.__doc__ = "Merge population"
    
    def __call__(self, parent_f: np.ndarray,\
                       offspring_f: np.ndarray,\
                       features: dict) -> dict:
        """
        ------------------------------------------------------
        Description:
            survivor selection algorithm, where choose a individuals between parents and 
            offspring that have great f for the next generation. 
        Arguments:
            parent_f: 1-D numpy array.
            offspring_f: 1-D numpy array.
            features: dictionary where every key is a list of two numpy arrays, which means 
                    key- [parent population feature , offspring population feature].
                Note: This features will be saved to the next generation.
            About:
                If you want more information, check:
                
        ------------------------------------------------------
        """  
        result = {}

        tmp_f = np.concatenate((parent_f,offspring_f))
        indices = get_lowest_indices(tmp_f, len(parent_f))
        result['parent_population_f'] = tmp_f[indices]

        for feature in features.keys():
            if len(features[feature]) != 2:
                raise Exception("Lenght of list have to be 2 (parent population feature and offspring population feature).")
            tmp = np.concatenate((features[feature][0], features[feature][1]), axis=0)
            result[feature] = deepcopy(tmp[indices])
        return result

class replacement_selector:
    def __init__(self):
        self.__doc__ = "Replacement population"

    def __call__(self, parent_f: np.ndarray,\
                       offspring_f: np.ndarray,\
                       features: dict) -> dict:
        """
        ------------------------------------------------------
        Description:
            survivor selection algorithm, where choose m individuals in 
            offspring that have great f for the next generation. 
        Arguments:
            parent_f: 1-D numpy array.
            offspring_f: 1-D numpy array.
            features: dictionary where every key is a list of two numpy arrays, which means 
                    key- [parent population feature , offspring population feature].
                Note: This features will be saved to the next generation.
            About:
                If you want more information, check:
                
        ------------------------------------------------------
        """  
        result = {}
        indices = get_lowest_indices(offspring_f, len(parent_f))
        
        result['parent_population_f'] = offspring_f[indices]

        for feature in features.keys():
            if len(features[feature]) != 2:
                raise Exception("Lenght of list have to be 2 (parent population feature and offspring population feature).")
            result[feature] = deepcopy(features[feature][1][indices])

        return result


import numpy as np 
from numba import jit,prange

__all__ = ['insertion_mutator','exchange_mutator','boundary_mutator','uniform_mutator',\
           'non_uniform_mutator','none_mutator','sigma_mutator','mult_sigma_adaptive_mutator',\
           'single_sigma_adaptive_mutator','sigma_ep_adaptive_mutator', 'sigma_ep_adaptive',\
            'single_sigma_adaptive', 'mult_sigma_adaptive', 'mutation_by_sigma', 'insertion_mutation',\
            'insertion_mutation', 'exchange_mutation', 'boundary_mutation', 'uniform_mutation',\
            'non_uniform_mutation']


"""
---------------------------------------------------------------------------------
                                Evolution Strategy.
---------------------------------------------------------------------------------
"""
#Mutation operators.
@jit(nopython=True, parallel=True)
def sigma_ep_adaptive(X: np.ndarray, alpha:float) -> np.ndarray:
    return X * (1.0 + alpha * np.random.normal(0,1,size=X.shape))

@jit(nopython=True,parallel=True)
def single_sigma_adaptive(X: np.ndarray,gamma:float) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Every individual has a sigma parameter which helps you to 
        update in a fixed portion the individual.
        
        Note:
            This function is for Evolution Strategy search. 
    Arguments: 
        - X: Numpy array where every N-th row is an individual and 
          M columns that are the decision variables.
        - gamma: Float value, additional parameter. 

        About:
            If you want more information, check:
            ¡INCLUIR ARTICULO SOBRE ESTO!
    ------------------------------------------------------ 
    """
    exponent = np.random.normal(0,1) * gamma
    return np.exp(exponent) * X

@jit(nopython=True, parallel=True)
def mult_sigma_adaptive(X: np.ndarray, gamma: float, gamma_prime: float) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Every individual has a sigma for every decision variable which helps you to 
        update in a fixed portion the individual.

        Note: 
            This function is for Evolution Strategy search.
    Arguments: 
        - X: Numpy array where every N-th row is an individual and 
          M columns that are the decision variables.
        - gamma: Float value, additional parameter. 
        - gamma_prime: Float value, additional parameter.

    About: 
        If you want more information, check:
        ¡INCLUIR ARTICULO SOBRE ESTO!
    ------------------------------------------------------ 
    """
    firts_ = np.random.normal(0,1) * gamma_prime 
    second_ = np.random.normal(0,1,size=X.shape) * gamma

    exponent = firts_ + second_
    return X* np.exp(exponent)   


@jit(nopython=True,parallel=False)
def mutation_by_sigma(X: np.ndarray, sigma: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Update function in the decision variables of every individual.

        Note: 
            This function is for Evolution Strategy search.
    Arguments: 
        - X: Numpy Matrix, where every N-th row is an individual and 
          M columns that are the decision variables.
        - sigma: Numpy Array with N rows which means every row is a individual, M columns differ in
        what kind of sigma has chosen (single or multiple).
    ------------------------------------------------------ 
    """
    NormalValue = np.random.normal(0,1)
    return  X +   sigma * NormalValue

"""
---------------------------------------------------------------------------------
                                Genetic Algortihms.
---------------------------------------------------------------------------------
"""

#Discrete mutation.
@jit(nopython=True, parallel=True)
def insertion_mutation(X : np.ndarray, n_elements: int=1) -> np.ndarray:
    num_individuals    = len(X)
    decision_variables = len(X[0])

    X_mutated = np.ones((num_individuals,decision_variables))

    for ind in prange(num_individuals):
        individual = np.full(decision_variables,np.inf)
        indices_elements = np.random.choice(decision_variables,n_elements, replace=False)
        elements = X[ind][indices_elements]
        indices_position = np.random.choice(decision_variables,n_elements, replace= False)
        individual[indices_position] = elements

        indices_elements.sort()
        remaining = []
        start_ = 0
        for e in range(decision_variables):
            if start_ < len(indices_elements) and e == indices_elements[start_]:
                start_+=1
                continue
            remaining.append(X[ind][e])

        individual[individual == np.inf] = remaining
        X_mutated[ind] = individual

    return X_mutated

@jit(nopython=True, parallel=True)
def exchange_mutation(X : np.ndarray) -> np.ndarray:
    num_individuals    = len(X)
    decision_variables = len(X[0])

    for ind in prange(num_individuals):
        exchange_points = np.random.choice(decision_variables,2,replace=False)
        x1 = exchange_points[0]
        x2 = exchange_points[1]

        tmp_point= X[ind][x1]
        X[ind][x1] = X[ind][x2]
        X[ind][x2] = tmp_point

    return X

#Continuos mutation.
@jit(nopython=False, parallel=True)
def boundary_mutation(X: np.ndarray, lower_bound: (list,int),\
                      upper_bound: (list,int)) -> np.ndarray:
    num_individuals = len(X)
    decision_variables = len(X[0])

    LB,UB = lower_bound, upper_bound
    for ind in prange(num_individuals):
        variable = np.random.randint(0,decision_variables)
        
        if type(lower_bound) == list:
            LB = lower_bound[variable]
            UB = upper_bound[variable]
                
        X[ind][variable] = LB

        if np.random.rand()>0.5:
            X[ind][variable] = UB
        
    return X

@jit(nopython=False, parallel=True)
def uniform_mutation(X: np.ndarray, lower_bound: (list,int),\
                     upper_bound: (list,int)) -> np.ndarray:
    num_individuals = len(X)
    decision_variables = len(X[0])

    LB,UB = lower_bound, upper_bound
    for ind in prange(num_individuals):
        variable = np.random.randint(0,decision_variables)

        if type(lower_bound) == list:
            LB = lower_bound[variable]
            UB = upper_bound[variable]    

        X[ind][variable] = np.random.uniform(LB, UB)
        
    return X

@jit(nopython=False, parallel=True)
def non_uniform_mutation(X:np.ndarray, sigma: float=1.0) -> np.ndarray:

    num_individuals    = len(X)
    decision_variables = len(X[0])

    for ind in prange(num_individuals):
        noise = np.random.normal(0,sigma, size= decision_variables)
        X[ind] *= noise

    return X

class insertion_mutator:
    def __init__(self, n_elements: int=1):
        self.n_elements = n_elements
        self.__doc__ = "Insertion \n\t Arguments:\n\t\t -n_elements: {}".format(n_elements)

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return insertion_mutation(X, self.n_elements)

class exchange_mutator:
    def __init__(self):
        self.__doc__ = "Exchange"

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return exchange_mutation(X)

class boundary_mutator:
    def __init__(self, bounds:list):
        self.Bounds = bounds
        self.__doc__="Boundary\n\t Arguments:\n\t\t -Lower bound: {}\n\t\t -Upper bound: {}".format(self.Bounds[0],self.Bounds[1])

    def __call__(self, X:np.ndarray) -> np.ndarray:
        return boundary_mutation(X, self.Bounds[0], self.Bounds[1])

class uniform_mutator:
    def __init__(self, bounds: list):
        self.Bounds = bounds
        self.__doc__ = "Uniform\n\t Arguments:\n\t\t -Lower bound: {}\n\t\t -Upper bound: {}".format(self.Bounds[0],self.Bounds[1])

    def __call__(self, X:np.ndarray) -> np.ndarray:
        return uniform_mutation(X, self.Bounds[0], self.Bounds[1])

class non_uniform_mutator:
    def __init__(self, sigma: float=1.0):
        self.sigma = sigma
        self.__doc__ = "Non Uniform\n\t Arguments:\n\t\t -Sigma: {}".format(self.sigma)

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return non_uniform_mutation(X,self.sigma)

class none_mutator:
    def __init__(self):
        self.__doc__ = "None"

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return X

#Mutatio operator for EP.
class sigma_ep_adaptive_mutator:
    def __init__(self,decision_variables:int ,alpha: float):
        self._alpha = alpha
        self._length = decision_variables
        self.__doc__ = "Sigma EP.\n\t\t-Alpha: {}".format(self._alpha)

    @property
    def length(self):
        return self._length

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return sigma_ep_adaptive(X,self._alpha)

#Mutation operator for ES.
class sigma_mutator:
    def __init__(self):
        self.__doc__ = "Sigma"

    def __call__(self, X: np.ndarray, Sigma: np.ndarray) -> np.ndarray:
        return mutation_by_sigma(X,Sigma)

class mult_sigma_adaptive_mutator:
    def __init__(self, decision_variables):
        self.__doc__ = "Sigma mult"
        self._length = decision_variables
        self._gamma_prime = 1/np.sqrt(2*self._length)
        self._gamma = 1/ np.sqrt( 2 * np.sqrt(self._length))

    @property
    def length(self):
        return self._length
    
    def __call__(self, sigma: np.ndarray) -> np.ndarray:
        return mult_sigma_adaptive(sigma,self._gamma, self._gamma_prime)

class single_sigma_adaptive_mutator:
    def __init__(self, decision_variables):
        self.__doc__ = "Single Sigma"
        self._length = 1
        self._tau = 1/ np.sqrt(self._length)

    @property
    def length(self):
        return self._length
    
    def __call__(self, sigma: np.ndarray) -> np.ndarray:
        return single_sigma_adaptive(sigma , self._length)


from pyristic.utils.helpers import *
from random import random 
import math
import numpy as np
import copy 

__all__=['SimulatedAnnealing']

class SimulatedAnnealing:

    def __init__(self, function : function_type,\
                       constraints : list):
        """
        ------------------------------------------------------
        Description:
            Initializing every variable necessary to the search.
        Arguments: 
            - Function: Objective function to minimize.
            - Constraints: Constraints to be a feasible solution.
        ------------------------------------------------------ 
        """
        self.f = function
        self.Constraints = constraints

        #Search information.
        self.logger = {}
        self.logger['best_individual']   = None
        self.logger['best_f']     = None

    def __str__(self):
        printable = "Simulated Annealing: \n f(X) = {} \n X = {} \n ".format(self.logger['best_f'], self.logger['best_individual'])
        first = True

        for i in range(len(self.Constraints)):
            if self.Constraints[i].__doc__ != None:

                if first:
                    first = False
                    printable += "Constraints: \n "

                self.Constraints[i](self.logger['best_individual'])
                printable += "{} \n".format( self.Constraints[i].__doc__)
        return printable
    
    
    def optimize(self,  Init : (np.ndarray,function_type),\
                        IniTemperature : float ,\
                        eps : float,\
                        **kwargs) -> None:
        """
        ------------------------------------------------------
        Description:
            Main function to find the best solution using Simulated Annealing strategy.
        Arguments:
            - Init: Numpy array, represent the initial solution or function which generates a random 
              initial solution (this solution should be a numpy array).
            - IniTemperature:  Floating value, it define how much allow worse solutions.
            - eps: it means what's the final temperature.
        ------------------------------------------------------
        """

        if type(Init) == function_type:
            candidate = Init()
        else:
            candidate = copy.deepcopy(Init)

        self.logger['best_individual'] = copy.deepcopy(candidate)
        self.logger['best_f'] = self.f(copy.deepcopy(candidate))
        self.logger['temperature'] = IniTemperature
        
        f_candidate =self.f(candidate)


        while self.logger['temperature'] >= eps:
            Neighbor = self.get_neighbor(candidate,**kwargs)
            
            if not self.is_valid(Neighbor):
                continue

            f_neighbor = self.f(Neighbor)

            if f_neighbor < f_candidate or random() < self.P(f_neighbor,f_candidate,**kwargs):
                candidate = Neighbor
                f_candidate = f_neighbor
            
            self.logger['temperature'] = self.update_temperature(**kwargs)
            
            #Update best solution found.
            if f_candidate < self.logger['best_f']:
                self.logger['best_f'] = f_candidate
                self.logger['best_individual'] = candidate

        
    def P(self, f_n : float , f_x : float,**kwargs) -> float:
        T = self.logger['temperature']
        return math.exp( -(f_n - f_x)/ T)

    def is_valid(self , x: np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is valid. 
        ------------------------------------------------------
        """
        for constrain in self.Constraints:
            if not constrain(x):
                return False
        return True
    
    def update_temperature(self,**kwargs) -> float:
        """
        ------------------------------------------------------
        Description:

        ------------------------------------------------------
        """
        return  self.logger['temperature'] *0.99

    def get_neighbor(self, x : np.ndarray,**kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
           Return only one solution which is a "random" 
            variation of current solution.
        ------------------------------------------------------
        """
        raise NotImplementedError





from pyristic.utils.operators import selection, mutation, crossover
from pyristic.utils.helpers import function_type, ContinuosFixer, EvolutionStrategyConfig
from tqdm import tqdm
import numpy as np

__all__= ['EvolutionStrategy']


class EvolutionStrategy:
    def __init__(self,  function: function_type,\
                        decision_variables:int,\
                        constraints:list=[],\
                        bounds: list=[],\
                        config = EvolutionStrategyConfig()):
        """
        ------------------------------------------------------
        Description:
            Initializing every variable necessary to the search.
        Arguments: 
            - Function: Objective function to minimize.
            - Constraints: Constraints to be a feasible solution.
            - Bounds: bound for every variable, this should be a matrix 2 x N
            where N is the variables number. The first element is lowert limit 
            and another one is the upper limit. 
        ------------------------------------------------------ 
        """

        self.f = function
        self.Constraints = constraints
        self.Bounds = bounds
        self.Decision_variables = decision_variables
        
        #Configuration.
        self._mutation_operator           = config.mutation_op  if config.mutation_op != None else mutation.sigma_mutator()
        self._crossover_operator          = config.cross_op if config.cross_op != None else crossover.discrete_crossover()
        self._survivor_selector           = config.survivor_selector if config.survivor_selector != None else selection.merge_selector()
        self._fixer                       = config.fixer if config.fixer != None else  ContinuosFixer(self.Bounds)
        self._adaptive_crossover_operator = config.adaptive_crossover_op if config.adaptive_crossover_op != None else crossover.intermediate_crossover()
        self._adaptive_mutation_operator  = config.adaptive_mutation_op  if config.adaptive_mutation_op != None else mutation.mult_sigma_adaptive_mutator(self.Decision_variables)

        #Search information.
        self.logger = {}
        self.logger['best_individual']      = None
        self.logger['best_f']        = None
        self.logger['current_iter']    = None
        self.logger['total_iter']      = None

    def __str__(self):
        printable = "Evolution Strategy search: \n f(X) = {} \n X = {} \n ".format(self.logger['best_f'], self.logger['best_individual'])
        first = True
        
        for i in range(len(self.Constraints)):
            if self.Constraints[i].__doc__ != None:
                
                if first:
                    first = False
                    printable += "Constraints: \n "
                    
                self.Constraints[i](self.logger['best_individual'])
                printable += "{} \n".format( self.Constraints[i].__doc__)
                    
        return printable

    def optimize(self,  generations:     int ,\
                        population_size: int,\
                        offspring_size:  int,\
                        eps_sigma:       float=0.001,\
                        verbose=True,\
                        **kwargs) -> None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using tabu search.
        Arguments:
            - generations: integer that represent the total iterations.
            - population_size: the population that 
        ------------------------------------------------------
        """
        #Reset global solution.
        self.logger['best_individual']      = None
        self.logger['best_f']        = None
        self.logger['current_iter']    = 0
        self.logger['total_iter']      = generations
        self.logger['parent_population_size'] = population_size
        self.logger['offspring_population_size']  = offspring_size       
        #offspring_size//=2
        self.logger['parent_population_x']= self.initialize_population(**kwargs)
        self.logger['parent_population_sigma'] = self.initialize_step_weights(eps_sigma,**kwargs)
        self.logger['parent_population_f'] = np.apply_along_axis( self.f ,\
                                                                        1,\
                                                                        self.logger['parent_population_x'])   

        try:
            for g in tqdm(range(generations), disable = not verbose):
                   

                #Crossover.
                first_parent_indices, second_parent_indices = self.get_pairs(**kwargs)
                self.logger['offspring_population_x']       = self.crossover_operator(  first_parent_indices,\
                                                                                        second_parent_indices,\
                                                                                        **kwargs)

                self.logger['offspring_population_sigma'] = self.adaptive_crossover(    first_parent_indices,\
                                                                                        second_parent_indices,\
                                                                                        **kwargs)
                #mutate.
                self.logger['offspring_population_sigma'] = self.adaptive_mutation(**kwargs)
                self.logger['offspring_population_x']     = self.mutation_operator(**kwargs)
                
                #Fixing solutions and getting aptitude.
                f_offspring = []
                for i in range(len(self.logger['offspring_population_x'])):
                    if self.is_invalid(self.logger['offspring_population_x'][i]):
                        self.logger['offspring_population_x'][i] = self.fixer(i)
                    f_offspring.append(self.f(self.logger['offspring_population_x'][i]))

                self.logger['offspring_population_f'] = np.array(f_offspring)
                
                next_generation = self.survivor_selection(**kwargs)
                self.logger['parent_population_x']        = next_generation['parent_population_x']
                self.logger['parent_population_sigma']    = next_generation['parent_population_sigma']
                self.logger['parent_population_f']  = next_generation['parent_population_f']
                self.logger['current_iter'] += 1

        except KeyboardInterrupt:
            print("Interrupted, saving best solution found so far.")
        
        ind = np.argmin(self.logger['parent_population_f'])

        self.logger['best_individual'] = self.logger['parent_population_x'][ind]
        self.logger['best_f']   = self.logger['parent_population_f'][ind]

    def initialize_step_weights(self, eps_sigma:float, **kwargs) -> np.ndarray :
        steps = np.random.uniform(0,1, size=(self.logger['parent_population_size'], self._adaptive_mutation_operator.length))
        return np.maximum(steps,eps_sigma)

    def initialize_population(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            How to distribute the population. This function create a
            population using uniform distribution.
            This should return an matrix of size (M x N)
            where M is the number of population and N is the number of 
            variables.
        Arguments:
            -size_: Tnteger n where n is the number of variables about the problem.
        ------------------------------------------------------
        """
        return np.random.uniform( self.Bounds[0], self.Bounds[1], size=(self.logger['parent_population_size'],self.Decision_variables))

    def fixer(self, ind:int) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
        Function which helps to move solution in a valid region.
        Arguments:
            - ind: index of individual.
        ------------------------------------------------------
        """
        return self._fixer(self.logger['offspring_population_x'], ind)
    
    def crossover_operator(self, parent_ind1: np.ndarray,\
                                 parent_ind2: np.ndarray,\
                                 **kwargs) -> np.ndarray:
        return self._crossover_operator(self.logger['parent_population_x'],\
                                        parent_ind1,\
                                        parent_ind2)

    def adaptive_crossover(self, parent_ind1: np.ndarray,\
                                 parent_ind2: np.ndarray,\
                                 **kwargs) -> np.ndarray:
        return self._adaptive_crossover_operator(   self.logger['parent_population_sigma'],\
                                                    parent_ind1,\
                                                    parent_ind2)

    def mutation_operator(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            The current population is updated by specific change
            using the adaptive control. 
            This function should mutate the population.
        ------------------------------------------------------
        """
        return  self._mutation_operator(self.logger['offspring_population_x'],\
                                        self.logger['offspring_population_sigma'])

    def adaptive_mutation(self, **kwargs) -> np.ndarray:
        return self._adaptive_mutation_operator(self.logger['offspring_population_sigma'])

    def survivor_selection(self,**kwargs) -> dict:
        individuals = {}
        individuals['parent_population_x']     = [self.logger['parent_population_x'],\
                                                  self.logger['offspring_population_x']]
        individuals['parent_population_sigma'] = [self.logger['parent_population_sigma'],\
                                                  self.logger['offspring_population_sigma']]

        return self._survivor_selector( self.logger['parent_population_f'],\
                                        self.logger['offspring_population_f'],\
                                        individuals)
    #-----------------------------------------------------
                    #Private functions.
    #-----------------------------------------------------
    def is_invalid(self, x : np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is invalid. 
        ------------------------------------------------------
        """
        for constraint in self.Constraints:
            if not constraint(x):
                return True
        return False
    
    def get_pairs(self, **kwargs ):
        parent_ind1 = np.random.randint(self.logger['parent_population_size'], size=(self.logger['offspring_population_size'],))
        parent_ind2 = np.random.randint(self.logger['parent_population_size'], size=(self.logger['offspring_population_size'],))
    
        return parent_ind1,parent_ind2



from pyristic.utils.helpers import function_type, ContinuosFixer, EvolutionaryProgrammingConfig
from pyristic.utils.operators import mutation,selection
from tqdm import tqdm
import numpy as np

__all__=['EvolutionaryProgramming']

class EvolutionaryProgramming:
    
    def __init__(self,  function: function_type,\
                        decision_variables: int,\
                        constraints:list=[],\
                        bounds: list=[],\
                        config = EvolutionaryProgrammingConfig()):
        """
        ------------------------------------------------------
        Description:
            Initializing every variable necessary to the search.
        Arguments: 
            - Function: Objective function to minimize.
            - Decision variables: Number of decision variables. 
            - Constraints: Constraints to be a feasible solution.
            - Bounds: Bounds for each variable. This should be a matrix 2 x N
            where N is the variables number. The first element is the lower limit, 
            and another one is the upper limit.  
            - Config: EvolutationaryProgrammingConfig, which changes the behavior of
            some operators.
        ------------------------------------------------------ 
        """

        self.f = function
        self.Constraints = constraints
        self.Bounds = bounds
        self.Decision_variables = decision_variables

        #Setting operators.
        self._mutation_operator           = config.mutation_op          if config.mutation_op != None          else mutation.sigma_mutator()
        self._survivor_selector           = config.survivor_selector    if config.survivor_selector != None    else selection.merge_selector()
        self._fixer                       = config.fixer                if config.fixer != None                else ContinuosFixer(self.Bounds)
        self._adaptive_mutation_operator  = config.adaptive_mutation_op if config.adaptive_mutation_op != None else mutation.sigma_ep_adaptive_mutator(self.Decision_variables, 0.5)

        #Global information.
        self.logger = {}
        self.logger['best_individual']              = None
        self.logger['best_f']                       = None
        self.logger['current_iter']                 = None
        self.logger['total_iter']                   = None
        self.logger['population_size']              = None 
        self.logger['parent_population_x']          = None
        self.logger['parent_population_sigma']      = None
        self.logger['parent_population_f']          = None   
        self.logger['offspring_population_x']       = None
        self.logger['offspring_population_sigma']   = None
        self.logger['offspring_population_f']       = None

    def __str__(self):
        printable = "Evolutive Programming search: \n f(X) = {} \n X = {} \n ".format(self.logger['best_f'], self.logger['best_individual'])
        first = True
        
        for i in range(len(self.Constraints)):
            if self.Constraints[i].__doc__ != None:
                
                if first:
                    first = False
                    printable += "Constraints: \n "
                    
                self.Constraints[i](self.logger['best_individual'])
                printable += "{} \n".format( self.Constraints[i].__doc__)
                    
        return printable

    def optimize(self,  generations: int ,\
                        size_population: int,\
                        verbose=True,\
                        **kwargs) ->None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using Evolutionary programming.
        Arguments:
            - Generations: Number of iterations.
            - Population size: Number of individuals. 
        ------------------------------------------------------
        """
        #Reset global solution.
        self.logger['current_iter']       = 0
        self.logger['total_iter']         = generations
        self.logger['population_size']    = size_population
        self.logger['best_individual']         = None
        self.logger['best_f']           = None  

        #Initial population.
        
        self.logger['parent_population_x']       = self.initialize_population(**kwargs)
        self.logger['parent_population_sigma']   = self.initialize_step_weights(**kwargs)
        self.logger['parent_population_f'] =  np.apply_along_axis(self.f , 1, self.logger['parent_population_x'])      



        try:
            for g in tqdm(range(generations), disable= not verbose):
                #Mutation.
                self.logger['offspring_population_sigma']  = self.adaptive_mutation(**kwargs)
                
                self.logger['offspring_population_x']      = self.mutation_operator(**kwargs)
                
                #Fixing solutions and getting aptitude.
                f_offspring = []
                for ind in range(len(self.logger['offspring_population_x'])):
                    if self.is_invalid(self.logger['offspring_population_x'][ind]):
                        self.logger['offspring_population_x'][ind] = self.fixer(ind)
                    f_offspring.append(self.f(self.logger['offspring_population_x'][ind]))
                self.logger['offspring_population_f'] = np.array(f_offspring)

                #Survivor selection.
                next_generation = self.survivor_selection(**kwargs)
                self.logger['parent_population_x']        = next_generation['parent_population_x']
                self.logger['parent_population_sigma']    = next_generation['parent_population_sigma']
                self.logger['parent_population_f']  = next_generation['parent_population_f']

                self.logger['current_iter']  += 1

        except KeyboardInterrupt:
            print("Interrupted, saving best solution found so far.")
        
        ind = np.argmin(self.logger['parent_population_f'])
        self.logger['best_individual'] = self.logger['parent_population_x'][ind]
        self.logger['best_f']   = self.logger['parent_population_f'][ind]
    
    def mutation_operator(self,**kwargs) -> None:
        """
        ------------------------------------------------------
        Description:
            A new population is created from the current population. 
            This function applies small changes to the decision variables 
            and the step sizes.
        ------------------------------------------------------
        """
        return self._mutation_operator(self.logger['parent_population_x'],\
                                       self.logger['parent_population_sigma'])

    def adaptive_mutation(self, **kwargs) -> None:
        return self._adaptive_mutation_operator(self.logger['parent_population_sigma'])

    def survivor_selection(self,**kwargs) -> dict:
        individuals = {}
        individuals['parent_population_x'] = [self.logger['parent_population_x'], self.logger['offspring_population_x']]
        individuals['parent_population_sigma']      = [self.logger['parent_population_sigma'], self.logger['offspring_population_sigma']]

        return self._survivor_selector( self.logger['parent_population_f'],\
                                        self.logger['offspring_population_f'],\
                                        individuals)

    def initialize_step_weights(self, **kwargs) -> np.ndarray :
        steps = np.random.uniform(0,1, size=(self.logger['population_size'] , self.Decision_variables))
        return np.maximum(steps,0.00001)

    def initialize_population(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            This function creates a population using a uniform distribution.
            It returns a matrix of size (n x m), where n is the size population 
            and m is the number of variables.
        Arguments:
            -n: The population size. 
        ------------------------------------------------------
        """
        return np.random.uniform( self.Bounds[0], self.Bounds[1], size=(self.logger['population_size'],self.Decision_variables))

    def fixer(self, ind:int) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
        This function helps to move one solution to the feasible region.
        Arguments:
            - ind: Index of the individual.
        ------------------------------------------------------
        """
        return  self._fixer(self.logger['offspring_population_x'], ind)

    def is_invalid(self, x : np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            This function checks if the current solution is unfeasible. 
        ------------------------------------------------------
        """
        for constraint in self.Constraints:
            if not constraint(x):
                return True
        return False


from numba import jit,prange
import numpy as np 
import math 


"""
--------------------------------------------------------------------------
                            Beale function.
--------------------------------------------------------------------------
"""
@jit(nopython=True, parallel=False)
def beale_function(X : np.ndarray) -> float:
    a = (1.5 - X[0] + X[0]*X[1])**2
    b = (2.25 - X[0] + X[0]*X[1]**2)**2
    c = (2.625 - X[0] + X[0]*X[1]**3)**2
    return a+b+c


def constraint1_beale(X: np.ndarray) -> bool:
    for i in range(len(X)):
        if -4.5>X[i] or X[i] > 4.5:
            return False 
    #If you want to see the result.
    constraint1_beale.__doc__="x1: -4.5 <= {:.2f} <= 4.5 \n x2: -4.5 <= {:.2f} <= 4.5".format(X[0],X[1])
    return True

beale_constraints = [constraint1_beale]
beale_bounds = [-4.5,4.5]

beale_ = {'function':beale_function, 'constraints': beale_constraints,\
          'bounds': beale_bounds,'decision_variables':2}

"""
--------------------------------------------------------------------------
                            Ackley function.
--------------------------------------------------------------------------
"""
@jit(nopython=True,parallel=False)
def ackley_function(X : np.ndarray) -> float:
    n = len(X)
    square_sum = (1/n)*np.sum(X * X)
    trigonometric_sum = (1/n)*np.sum(np.cos(2*np.pi*X))
 
    return -20*np.exp( -0.2 * np.sqrt(square_sum)) - np.exp(trigonometric_sum) + 20 + np.e

def constraint1_ackley(X : np.ndarray) -> bool:
    str_ = ""
    valid = True
    for i in range(len(X)):
        if -30>X[i] or X[i] > 30:
            valid = False
        str_+= "x{}: -30 <= {:.2f} <= 30 \n ".format(i+1, X[i])
    
    #Important if you want to see the result.    
    constraint1_ackley.__doc__= str_
    return valid

ackley_constraints = [constraint1_ackley]
ackley_bounds = [-30.0,30.0]

ackley_ = {'function': ackley_function, 'constraints': ackley_constraints,\
           'bounds': ackley_bounds, 'decision_variables': 10}

"""
--------------------------------------------------------------------------
                            Bukin N.6 function.
--------------------------------------------------------------------------
"""

@jit(nopython=True, parallel=True)
def bukin_function(X: np.ndarray) -> float:
    A = np.sqrt( np.abs(X[1] - 0.01* X[0]**2) ) 
    B = np.abs(X[0] + 10)
    return 100.0 * A + 0.01* B

def constraint1_bukin(X : np.ndarray) -> bool:
    str_ = ""
    str_+= "x{}: -15 <= {:.2f} <= -5 \n ".format(1, X[0])    
    str_+= "x{}: -3 <= {:.2f} <= 3 \n ".format(2, X[1])
    constraint1_bukin.__doc__= str
    if -15>X[0] or X[0] > -5:
        return False
    if -3 > X[1] or X[1] > 3:
        return False
    return True

bukin_constraints = [constraint1_bukin]
bukin_bounds = [[-15,-3],[-5,3]] # [[Lower bounds] , [Upper bounds]]

bukin_ = {'function': bukin_function, 'constraints': bukin_constraints,\
           'bounds': bukin_bounds, 'decision_variables': 2}

"""
--------------------------------------------------------------------------
                            Himmelblau's function.
--------------------------------------------------------------------------
"""


@jit(nopython=True, parallel=True)
def Himmelblau_function(X: np.ndarray) -> float:
    A = (X[0]**2 + X[1] - 11)**2
    B = (X[0] + X[1]**2 - 7)**2
    return A+B

def constraint1_Himmelblau(X : np.ndarray) -> bool:
    str_ = ""
    valid = True
    for i in range(len(X)):
        if -5>X[i] or X[i] > 5:
            valid = False
        str_+= "x{}: -5 <= {:.2f} <= 5 \n ".format(i+1, X[i])
    
    #Important if you want to see the result.    
    constraint1_ackley.__doc__= str_
    return valid

Himmelblau_constraints = [constraint1_Himmelblau]
Himmelblau_bounds = [-5,5]

Himmelblau_ = {'function': Himmelblau_function, 'constraints': Himmelblau_constraints,\
           'bounds': Himmelblau_bounds, 'decision_variables': 2}


from pyristic.utils.helpers import function_type
from tqdm import tqdm
import numpy as np

__all__= ['Genetic']

class Genetic:

    def __init__(self,  function: function_type,\
                        decision_variables:int,\
                        constraints:list=[],\
                        bounds: list=[],\
                        config = None):
        """
        ------------------------------------------------------
        Description:
            Initializing every variable necessary to the search.
        Arguments:
            - Function: Objective function to minimize.
            - Constraints: Constraints to be a feasible solution.
            - Bounds: bound for every variable, this should be a matrix 2 x N
            where N is the variables number. The first element is lowert limit
            and another one is the upper limit.
        ------------------------------------------------------
        """
        #Information about problem.
        self.f = function
        self.Constraints = constraints
        self.Bounds = bounds
        self.Decision_variables = decision_variables # Decision variables.

        #Operators.
        if config != None:
            self._mutation_operator  = config.mutation_op
            self._crossover_operator = config.cross_op
            self._survivor_selector  = config.survivor_selector
            self._parent_selector    = config.parent_selector
            self._fixer              = config.fixer
        #Search information.
        self.logger = {}
        self.logger['best_individual']      = None
        self.logger['best_f']        = None
        self.logger['current_iter']    = None
        self.logger['total_iter']      = None
        self.logger['population_size'] = None

    def __str__(self):
        printable = "Genetic search: \n f(X) = {} \n X = {} \n ".format(self.logger['best_f'], self.logger['best_individual'])
        first = True

        for i in range(len(self.Constraints)):
            if self.Constraints[i].__doc__ != None:

                if first:
                    first = False
                    printable += "Constraints: \n "

                self.Constraints[i](self.logger['best_individual'])
                printable += "{} \n".format( self.Constraints[i].__doc__)

        return printable

    #-----------------------------------------------------
                    #Public functions.
    #-----------------------------------------------------
    def optimize(self,  generations:int ,\
                        size_population: int,\
                        verbose:bool=True,\
                        **kwargs) -> None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using tabu search.
        Arguments:
            -generations: number of iterations.
            -population: Number of new solutions.
        ------------------------------------------------------
        """
        #Reset global solution.
        self.logger['current_iter']    = 0
        self.logger['total_iter']      = generations
        self.logger['population_size'] = size_population
        self.logger['best_f']        = None
        self.logger['best_individual']      = None

        #Initial population.
        self.logger['parent_population_x']       = self.initialize_population(**kwargs)
        self.logger['parent_population_f'] = np.apply_along_axis(self.f , 1, self.logger['parent_population_x'])
        
        try:
            for g in tqdm(range(generations), disable = not verbose):
                #Parent selection.
                parent_ind = self.parent_selection(**kwargs)
                first_parent_indices, second_parent_indices = self.get_pairs(parent_ind)
                
                #Crossover.
                self.logger['offspring_population_x'] = self.crossover_operator(first_parent_indices,\
                                                                                second_parent_indices,\
                                                                                **kwargs)
                #mutate.
                self.logger['offspring_population_x'] = self.mutation_operator(**kwargs)

                #Fixing solutions and getting aptitude.
                f_offspring = []
                for ind in range(len(self.logger['offspring_population_x'])):
                    if self.is_invalid(self.logger['offspring_population_x'][ind]):
                        self.logger['offspring_population_x'][ind] = self.fixer(ind)
                    f_offspring.append(self.f(self.logger['offspring_population_x'][ind]))
                self.logger['offspring_population_f'] = np.array(f_offspring)

                #Survivor selection.
                next_generation = self.survivor_selection(**kwargs)
                self.logger['parent_population_x']  = next_generation['population']
                self.logger['parent_population_f']  = next_generation['parent_population_f']

                self.logger['current_iter'] += 1

        except KeyboardInterrupt:
            print("Interrupted, saving best solution found so far.")

        ind = np.argmin(self.logger['parent_population_f'])
        self.logger['best_individual'] = self.logger['parent_population_x'][ind]
        self.logger['best_f']   = self.logger['parent_population_f'][ind]
    
    def initialize_population(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            How to distribute the population. This function create a
            population using uniform distribution.
            This should return an matrix of size (M x N)
            where M is the number of population and N is the number of
            variables.
        Arguments:
            -size_: tuple with two integers (m,n)
            where m is the number generated of new solutions and
            n is the number of variables about the problem.
        ------------------------------------------------------
        """
        return np.random.uniform( self.Bounds[0], self.Bounds[1], size=(self.logger['population_size'], self.Decision_variables))
        
    def fixer(self, ind: int) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
        Function which helps to move solution in a valid region.
        Arguments:
            -X: numpy array, this one is the current population.
            -ind: index of individual.
        ------------------------------------------------------
        """
        if self._fixer == None:
            raise NotImplementedError
        return self._fixer(self.logger['offspring_population_x'], ind)

    def mutation_operator(self, **kwargs):
        """
        ------------------------------------------------------
        Description:
            The current population is updated by specific change
            using step_size.
            This function should change the population and current step size
            for the next generation.
        Arguments:
            -Population: Matrix with size m x n, where m is the current size of
            population and n is the problem variables. Every row is an element.
            - step_size: Matrix with size m x n, where m is the current size of
            population and n is the poblem variables. Every row is an step for
            every element.
        ------------------------------------------------------
        """
        if self._mutation_operator == None:
            raise NotImplementedError
        return self._mutation_operator(self.logger['offspring_population_x'])

    def crossover_operator(self,parent_ind1: np.ndarray,\
                                parent_ind2: np.ndarray,\
                                **kwargs):
        if self._crossover_operator == None:
            raise NotImplementedError
        return self._crossover_operator(self.logger['parent_population_x'], parent_ind1, parent_ind2)

    def survivor_selection(self,**kwargs) -> dict:
        if self._survivor_selector == None:
            raise NotImplementedError

        individuals = {}
        individuals['population'] = [self.logger['parent_population_x'], self.logger['offspring_population_x']]

        return self._survivor_selector( self.logger['parent_population_f'],\
                                        self.logger['offspring_population_f'],\
                                        individuals)

    def parent_selection(self,**kwargs) -> np.ndarray:
        if self._parent_selector == None:
            raise NotImplementedError
        return self._parent_selector(self.logger['parent_population_f'])

    #-----------------------------------------------------
                    #Private functions.
    #-----------------------------------------------------
    def is_invalid(self, x : np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is invalid.
        ------------------------------------------------------
        """
        for constraint in self.Constraints:
            if not constraint(x):
                return True
        return False

    def get_pairs(self,parent_ind : np.ndarray):
        parent_ind1=[]
        parent_ind2=[]
        size_ = len(parent_ind)//2

        for a,b in np.random.randint(0,len(parent_ind),size=(size_,2)):
            parent_ind1.append(parent_ind[a])
            parent_ind2.append(parent_ind[b])

        return parent_ind1,parent_ind2



from pyristic.utils.helpers import *
from tqdm import tqdm
import numpy as np
import copy 

__all__=['TabuList','TabuSearch']

class TabuList:
    def __init__(self, timer: int=2):
        self._TB = [[None,None,-1, 100000000]]
        self.timer = timer
        
    def __str__(self):
        printable_object = "\n ---- Tabu list: ---- \n"
        printable_object += "List size: {} \n".format(len(self._TB))
        for i in range(len(self._TB)):
            a,b,c,d = self._TB[i][0], self._TB[i][1], self._TB[i][2], self._TB[i][3]
            printable_object += \
            "p: {} v:{} Iteration: {} current timer: {} \n".format(a,b,c,d)
        return printable_object
        
    def push(self, x: list) -> None:
        assert len(x) == 3
        x.append(self.timer)
        self._TB.insert(0,x)
    
    def pop_back(self) -> None:
        x_last = self._TB.pop()
        x_current_back = self.get_back()
        update_time = self.timer - (x_current_back[2]-x_last[2])
        x_current_back[2] -= update_time
        self.update_back(x_current_back)
        
    def get_back(self) -> list:
        return self._TB[-1]
    
    def update_back(self, x : list) -> None:
        self._TB[-1] = x

    def find(self, x : list) -> bool:
        #X is [p, v], where p is the position changed and v the new value.
        assert len(x) == 2
        for i in range(len(self._TB)):
            if x[0] == self._TB[i][0] and x[1] == self._TB[i][1]:
                return True
        return False

    def reset(self, timer : int) -> None:
        self._TB = [] 
        self.timer = timer
    
    def update(self) -> None:
        most_old_candidate = copy.deepcopy(self.get_back())
        most_old_candidate[-1] -= 1
           
        if(most_old_candidate[-1] <= 0): 
            self.pop_back()
        else:
            self.update_back(most_old_candidate)

class TabuSearch:
 
    def __init__(self,  function: function_type,\
                        constraints : list,\
                        TabuStruct = TabuList()):
        """
        Arguments: 
            - Function: Objective function to minimize.
            - Constraints: Constraints to be a feasible solution.
            - TabuStruct: Structure to store information about search by default is TabuList. 
        Note: 
            The argument in TabuList is not essential. It is changed when you call optimize.
        """
        self.TL = TabuStruct #Initialize tabulist
        self.Constraints = constraints
        self.f = function
        
        #Search information.
        self.logger = {}
        self.logger['best_individual']   = None
        self.logger['best_f']     = None
        self.logger['current_iter'] = None
        self.logger['total_iter']   = None
           
    def __str__(self):
        printable = "Tabu search: \n f(X) = {} \n X = {} \n ".format(self.logger['best_f'], self.logger['best_individual'])
        first = True
        
        for i in range(len(self.Constraints)):
            if self.Constraints[i].__doc__ != None:
                
                if first:
                    first = False
                    printable += "Constraints: \n "
                    
                self.Constraints[i](self.logger['best_individual'])
                printable += "{} \n".format( self.Constraints[i].__doc__)
                   
        return printable
        
    def optimize(self,Init: (np.ndarray,function_type) ,iterations: int,\
                 memory_time : int, **kwargs)->None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using tabu search.
        Arguments:
            - Init: Numpy array, represent the initial solution or function which generates a random 
              initial solution (this solution should be a numpy array).
            - Iterations:  Integer, stop condition.
            - Memory time: Integer, time which a solution is in the tabu list
        ------------------------------------------------------
        """
        self.TL.reset(memory_time)
        
        self.logger['best_individual'] = None
        self.logger['best_f']   = None
        
        
        if type(Init) == function_type:
            best_candidate = Init()
        else:
            best_candidate = copy.deepcopy(Init)
        
        f_candidate =self.f(best_candidate)
        
        try:
            for step_ in range(1,iterations+1):

                Neighbors =[neighbor for  neighbor in self.get_neighbors(best_candidate,**kwargs) \
                            if not self.TL.find(self.encode_change(neighbor, best_candidate,**kwargs))]
                Neighbors = np.array(Neighbors)

                ValidNeighbors =  \
                Neighbors[np.apply_along_axis(self.is_valid,1,Neighbors),:] 

                #Check if there exists a valid neighbor
                if len(ValidNeighbors) == 0:
                    continue

                f_feasible_candidates =  \
                np.apply_along_axis(self.f , 1, ValidNeighbors)      

                ind_min = np.argmin(f_feasible_candidates)           

                p,v = self.encode_change(ValidNeighbors[ind_min],best_candidate,**kwargs)
                self.TL.push([p, v, step_])

                if f_feasible_candidates[ind_min]< f_candidate:
                    best_candidate = copy.deepcopy(ValidNeighbors[ind_min])
                    f_candidate = f_feasible_candidates[ind_min]

                self.TL.update()

        except KeyboardInterrupt:
            print("Interrupted, saving best solution found so far.")
            
        self.logger['best_individual'] = best_candidate
        self.logger['best_f'] = f_candidate
    
    def is_valid(self, x : np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is valid. 
        ------------------------------------------------------
        """
        for constraint in self.Constraints:
            if not constraint(x):
                return False
        return True
    
    def get_neighbors(self, x : (list,np.ndarray),\
                            **kwargs)-> list:
        """
        ------------------------------------------------------
        Description:
            The user has to custom this function. 
            The function has to return a list with all possible neighbors. 
            Check the examples for a detailed explanation.
        ------------------------------------------------------
        """
        raise NotImplementedError
            
    def encode_change(self, neighbor: (list,np.ndarray),\
                            x: (list,np.ndarray),\
                            **kwargs) -> list:
        """
        ------------------------------------------------------
        Description: 
            The user has to custom this function.
            Check the examples for a detailed explanation.
        ------------------------------------------------------
        """
        raise NotImplementedError
     


# from distutils.core import setup
from setuptools import find_packages,setup

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
  name = 'pyristic',\
  version = '0.1.0',\
  license='MIT',\
  description = 'Set of metaheuristic for solve optimization problems.',\
  author = 'Jesús Armando Ortíz Peñafiel',\
  author_email = 'armandopenafiel12@gmail.com',\
  long_description = LONG_DESCRIPTION,\
  long_description_content_type="text/markdown",\
  url = 'https://github.com/JAOP1/',\
  download_url = 'https://github.com/JAOP1/pyristic/archive/refs/tags/0.1.0.tar.gz',\
  keywords = ['Optimization', 'Metaheuristic', 'Python'],\
  install_requires=[
          'numpy>=1.15.4',
          'tqdm>=4.28.1',
          'numba>=0.50.1'
      ],\
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7'
  ],\
  packages=find_packages(exclude=("examples",))
)


import numpy as np 
from numba import jit,prange

__all__ = ['intermediate_crossover','n_point_crossover',\
           'uniform_crossover','permutation_order_crossover','simulated_binary_crossover',\
            'discrete_cross', 'intermediate_cross', 'n_point_cross', 'uniform_cross', 'permutation_order_cross',\
            'simulated_binary_cross']

"""
---------------------------------------------------------------------------------
                                Evolution Strategy.
---------------------------------------------------------------------------------
"""

#Crossover operators.
@jit(nopython=False,parallel=True)
def discrete_cross(X : np.ndarray, parent_ind1: np.ndarray ,\
                       parent_ind2: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two elements.
        X_{i,s} if v == 1 
        X_{i,t} if v == 0
    Arguments:
        -X: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
    ------------------------------------------------------
    """

    #print(X)
    rows,cols = X.shape
    rows_new = len(parent_ind1)
    A = np.zeros((rows_new,cols))
    B = np.zeros((rows_new,cols))

    for r in prange(rows_new):
        for c in prange(cols):
            if np.random.randint(2):
                A[r,c] = 1
            else:
                B[r,c] = 1

    return X[parent_ind1] * A + X[parent_ind2] * B

@jit(nopython=False,parallel=True)
def intermediate_cross(X:np.ndarray, parent_ind1: np.ndarray,\
                       parent_ind2: np.ndarray, alpha: float=0.5) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining  the average of two elements.
    Arguments:
        -X: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
    ------------------------------------------------------
    """
    return alpha*X[parent_ind1]+ (1-alpha) * X[parent_ind2]

"""
---------------------------------------------------------------------------------
                                Genetic Algorithm.
---------------------------------------------------------------------------------
"""

@jit(nopython=True,parallel=True)
def n_point_cross(X : np.ndarray , parent_ind1 : np.ndarray,\
                  parent_ind2 : np.ndarray, n_cross: int = 1) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -X: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        -n_cross: integer value which says you the number of cuts (by default is 1).

        Note: 
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """
    assert len(parent_ind1) == len(parent_ind2)

    num_individuals = len(parent_ind1)
    decision_variables = len(X[0])    
    new_population = np.ones((num_individuals*2,decision_variables))
    new_population[np.arange(0,num_individuals*2,2)] = X[parent_ind1]
    new_population[np.arange(1,num_individuals*2,2)] = X[parent_ind2]

    for row in prange(num_individuals):
        cross_points  = np.random.choice(decision_variables-1,n_cross,replace=False)
        cross_points += 1
        cross_points.sort()
        for i in range(0,n_cross,2):
            start = cross_points[i]
            final = decision_variables 
            if i+1 < n_cross:
                final = cross_points[i+1]
            new_population[2*row][start:final]    = X[parent_ind2[row]][start:final]
            new_population[2*row +1][start:final] = X[parent_ind1[row]][start:final]

    return new_population

@jit(nopython=True, parallel=True)
def uniform_cross(X: np.ndarray, parent_ind1: np.ndarray,\
                  parent_ind2 : np.ndarray, flip_prob : int=0.5) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -X: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        -flip_prob: float value which says the likely to change a position (by default is 0.5).
        Note: 
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """
    assert len(parent_ind1) == len(parent_ind2)

    num_individuals = len(parent_ind1)
    decision_variables = len(X[0])
    new_population = np.ones((num_individuals*2,decision_variables))
    new_population[np.arange(0,num_individuals*2,2)] = X[parent_ind1]
    new_population[np.arange(1,num_individuals*2,2)] = X[parent_ind2]

    for i in prange(num_individuals):
        for x in range(decision_variables):
            if np.random.rand() <= flip_prob:
                new_population[2*i][x] = X[parent_ind2[i]][x]
                new_population[2*i+1][x] = X[parent_ind1[i]][x]

    return new_population

@jit(nopython=False, parallel=True)
def create_child(parent1: np.ndarray , parent2: np.ndarray) -> np.ndarray:

    """
        ¡¡¡FUNCTION OF permutation_order_cross!!!
    """
    decision_variables = len(parent1)
    interval = np.random.choice(decision_variables+1,2,replace=False)
    interval.sort()

    individual = np.full(decision_variables,np.inf)
    segment = parent1[interval[0]:interval[1]]
    individual[interval[0]:interval[1]] = segment

    remainder = []
    for x in parent2:
        if not x in segment:
            remainder.append(x)

    individual[individual == np.inf] = remainder
    return individual

@jit(nopython=False, parallel=True)
def permutation_order_cross(X : np.ndarray , parent_ind1:np.ndarray,\
                            parent_ind2:np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -X: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        Note: 
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """
    assert len(parent_ind1) == len(parent_ind2)

    num_individuals = len(parent_ind1)
    decision_variables = len(X[0])
    new_population = np.ones((num_individuals*2,decision_variables))

    for i in prange(num_individuals):
        new_population[2*i]  = create_child(X[parent_ind1[i]], X[parent_ind2[i]])
        new_population[2*i+1] = create_child(X[parent_ind2[i]], X[parent_ind1[i]])
    
    return new_population

@jit(nopython=True, parallel=True)
def simulated_binary_cross(X: np.ndarray, parent_ind1: np.ndarray,\
                           parent_ind2: np.ndarray,\
                           nc: int=1) -> np.ndarray:
    #assert len(parent_ind1) == len(parent_ind2)

    num_individuals = len(parent_ind1)
    decision_variables = len(X[0])                       

    u = np.random.rand()
    B = -1
    exponent = 1/(nc+1)
    if u <= 0.5:
        B = np.power((2*u), exponent)
    else:
        B = np.power(1/(2 * (1-u)), exponent)
    
    A = X[parent_ind1] + X[parent_ind2]
    B = np.absolute(X[parent_ind2] - X[parent_ind1]) * B

    return np.concatenate((0.5 * (A - B) , 0.5 * (A + B)) , axis = 0)


class intermediate_crossover:
    def __init__(self, alpha: float=0.5):
        self.alpha = alpha
        self.__doc__ = "Intermediate\n\tArguments:\n\t\t-Alpha:{}".format(self.alpha)
    
    def __call__(self,  population:np.ndarray,\
                        parent_ind1: np.ndarray,\
                        parent_ind2: np.ndarray) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return intermediate_cross(population,parent_ind1,\
                                parent_ind2, self.alpha)

class n_point_crossover:
    def __init__(self, n_cross: int = 1):
        self.n_cross = n_cross
        self.__doc__ = "n point\n\tArguments:\n\t\t-n_cross: {}".format(self.n_cross)

    def __call__(self,  population:np.ndarray,\
                    parent_ind1: np.ndarray,\
                    parent_ind2: np.ndarray) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return n_point_cross(population, parent_ind1, parent_ind2, self.n_cross)

class uniform_crossover:
    def __init__(self,flip_prob : int=0.5):
        self.prob = flip_prob
        self.__doc__ = "Uniform\n\tArguments:\n\t\t-prob: {}".format(self.prob)

    def __call__(self,  population:np.ndarray,\
                    parent_ind1: np.ndarray,\
                    parent_ind2: np.ndarray) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)    
        return uniform_cross(population, parent_ind1, parent_ind2, self.prob)

class permutation_order_crossover:
    def __init__(self):
        self.__doc__ = "Permutation order"

    def __call__(self,  population:np.ndarray,\
                    parent_ind1: np.ndarray,\
                    parent_ind2: np.ndarray) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return permutation_order_cross(population, parent_ind1, parent_ind2)

class simulated_binary_crossover:
    def __init__(self,nc: int=1):
        self.nc = nc
        self.__doc__ = "Simulated binary\n\tArguments:\n\t\t-nc: {}".format(self.nc)

    def __call__(self,  population: np.ndarray,\
                parent_ind1: np.ndarray,\
                parent_ind2: np.ndarray) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return simulated_binary_cross(population, parent_ind1,parent_ind2, self.nc)

class none_cross_crossover:
    def __init__(self):
        self.__doc__ = "None"

    def __call__(self, population : np.ndarray,\
                       parent_ind1: np.ndarray,\
                       parent_ind2: np.ndarray) -> np.ndarray:
        return population

#Crossover operator for ES.
class discrete_crossover:
    def __init__(self):
        self.__doc__ = "Discrete"
    
    def __call__(self, population : np.ndarray,\
                       parent_ind1: np.ndarray,\
                       parent_ind2: np.ndarray) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return discrete_cross(population, parent_ind1, parent_ind2)

    